from __future__ import absolute_import, division, print_function, unicode_literals

from amaasutils.random_utils import random_string
import random

from amaascore.books.book import Book


def generate_book(asset_manager_id=None, book_id=None, owner_id=None, party_id=None, book_type='Trading',
                  business_unit=None):

    book = Book(asset_manager_id=asset_manager_id or random.randint(1, 1000),
                book_id=book_id or random_string(10),
                book_type=book_type,
                owner_id=owner_id or random.randint(1, 1000),
                base_currency=random.choice(['USD', 'SGD', 'HKD', 'EUR']),
                business_unit=business_unit or random.choice(['Equities', 'Emerging Markets', 'Treasury']),
                party_id=party_id or random_string(10))  # Can also be the asset_manager_id of the book

    return book



from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import json

from amaascore.config import ENVIRONMENT
from amaascore.core.amaas_model import json_handler
from amaascore.core.interface import Interface
from amaascore.transactions.utils import json_to_transaction, json_to_position, json_to_mtm_result


class TransactionsInterface(Interface):

    def __init__(self, environment=ENVIRONMENT, logger=None, endpoint=None, username=None, password=None):
        self.logger = logger or logging.getLogger(__name__)
        super(TransactionsInterface, self).__init__(endpoint=endpoint, endpoint_type='transactions',
                                                    environment=environment, username=None, password=None)

    def new(self, transaction):
        self.logger.info('New Transaction - Asset Manager: %s - Transaction ID: %s', transaction.asset_manager_id,
                         transaction.transaction_id)
        url = '%s/transactions/%s' % (self.endpoint, transaction.asset_manager_id)
        response = self.session.post(url, json=transaction.to_interface())
        if response.ok:
            transaction = json_to_transaction(response.json())
            return transaction
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def create_many(self, transactions):
        if type(transactions) is not list:
            raise ValueError('Error - create_many takes in a list of transactions instead of single transaction')
        transactions_json = []        
        # check to ensure all transactions have the same asset_manager_id
        asset_manager_id = transactions[0].asset_manager_id
        for transaction in transactions:
            transactions_json.append(transaction.to_interface())
            if transaction.asset_manager_id != asset_manager_id:
                raise AttributeError('Check failed - Not all transactions have the same asset manager ID.')
        self.logger.info('Multple new Transactions - Asset Manager: %s', asset_manager_id)
        url = '%s/transactions/%s' % (self.endpoint, asset_manager_id)
        response = self.session.post(url, json=transactions_json)
        if response.ok:
            transactions = []
            for transaction_json in response.json():
                transactions.append(json_to_transaction(transaction_json))
            return transactions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def amend(self, transaction):
        self.logger.info('Amend Transaction - Asset Manager: %s - Transaction ID: %s', transaction.asset_manager_id,
                         transaction.transaction_id)
        url = '%s/transactions/%s/%s' % (self.endpoint, transaction.asset_manager_id, transaction.transaction_id)
        response = self.session.put(url, json=transaction.to_interface())
        if response.ok:
            transaction = json_to_transaction(response.json())
            return transaction
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def partial(self, asset_manager_id, transaction_id, updates):
        self.logger.info('Partial Amend Transaction - Asset Manager: %s - Transaction ID: %s', asset_manager_id,
                         transaction_id)
        url = '%s/transactions/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        response = self.session.patch(url, data=json.dumps(updates, default=json_handler), headers=self.json_header)
        if response.ok:
            transaction = json_to_transaction(response.json())
            return transaction
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve(self, asset_manager_id, transaction_id, version=None):
        self.logger.info('Retrieve Transaction - Asset Manager: %s - Transaction ID: %s', asset_manager_id,
                         transaction_id)
        url = '%s/transactions/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        if version:
            url += '?version=%d' % int(version)
        response = self.session.get(url)
        if response.ok:
            return json_to_transaction(response.json())
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def transactions_by_asset_manager(self, asset_manager_id):
        self.logger.info('Retrieve Transactions by Asset Manager: %s', asset_manager_id)
        url = '%s/transactions/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url)
        if response.ok:
            transactions = [json_to_transaction(json_transaction) for json_transaction in response.json()]
            self.logger.info('Returned %s Transactions.', len(transactions))
            return transactions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def cancel(self, asset_manager_id, transaction_id):
        self.logger.info('Cancel Transaction - Asset Manager: %s - Transaction ID: %s', asset_manager_id,
                         transaction_id)
        url = '%s/transactions/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        response = self.session.delete(url)
        if response.ok:
            self.logger.info('Successfully Cancelled - Asset Manager: %s - Transaction ID: %s.', asset_manager_id,
                             transaction_id)
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def search(self, asset_manager_id, transaction_ids=[], transaction_statuses=[],
               asset_book_ids=[], counterparty_book_ids=[], asset_ids=[], transaction_date_start=None,
               transaction_date_end=None, code_types=[], code_values=[], link_types=[], linked_transaction_ids=[],
               party_types=[], party_ids=[], reference_types=[], reference_values=[], client_ids=[],
               page_no=None, page_size=None):
        self.logger.info('Search Transactions - Asset Manager: %s', asset_manager_id)
        search_params = {}
        # Potentially roll this into a loop through args rather than explicitly named - depends on additional validation
        if transaction_ids:
            search_params['transaction_ids'] = ','.join(transaction_ids)
        if transaction_statuses:
            search_params['transaction_statuses'] = ','.join(transaction_statuses)
        if asset_book_ids:
            search_params['asset_book_ids'] = ','.join(asset_book_ids)
        if counterparty_book_ids:
            search_params['counterparty_book_ids'] = ','.join(counterparty_book_ids)
        if asset_ids:
            search_params['asset_ids'] = ','.join(asset_ids)
        if transaction_date_start:
            search_params['transaction_date_start'] = transaction_date_start
        if transaction_date_end:
            search_params['transaction_date_end'] = transaction_date_end
        if code_types:
            search_params['code_types'] = ','.join(code_types)
        if code_values:
            search_params['code_values'] = ','.join(code_values)
        if link_types:
            search_params['link_types'] = ','.join(link_types)
        if linked_transaction_ids:
            search_params['linked_transaction_ids'] = ','.join(linked_transaction_ids)
        if party_types:
            search_params['party_types'] = ','.join(party_types)
        if party_ids:
            search_params['party_ids'] = ','.join(party_ids)
        if reference_types:
            search_params['reference_types'] = ','.join(reference_types)
        if reference_values:
            search_params['reference_values'] = ','.join(reference_values)
        if client_ids:
            search_params['client_ids'] = ','.join(client_ids)
        if page_no is not None:
            search_params['page_no'] = page_no
        if page_size:
            search_params['page_size'] = page_size
        url = '%s/transactions/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url, params=search_params)
        if response.ok:
            transactions = [json_to_transaction(json_transaction) for json_transaction in response.json()]
            self.logger.info('Returned %s Transactions.', len(transactions))
            return transactions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def new_mtm_result(self, asset_manager_id, mtm_results):
        self.logger.info('Marking to market Positions - Asset Manager: %s', asset_manager_id)
        if not isinstance(mtm_results, list):
            mtm_results = [mtm_results]
        mtm_result_json = []
        for mtm_result in mtm_results:
            mtm_result_json.append(mtm_result.to_interface())
        url = '%s/mtm/%s' % (self.endpoint, asset_manager_id)
        response = self.session.post(url, json=mtm_result_json)
        if response.ok:
            mtm_results = []
            for mtm_result_json in response.json():
                mtm_results.append(json_to_mtm_result(mtm_result_json))
            return mtm_results
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def amend_mtm_result(self, asset_manager_id, mtm_results):
        self.logger.info('Amending mtm Positions - Asset Manager: %s', asset_manager_id)
        if not isinstance(mtm_results, list):
            mtm_results = [mtm_results]
        mtm_result_json = []
        for mtm_result in mtm_results:
            mtm_result_json.append(mtm_result.to_interface())
        url = '%s/mtm/%s' % (self.endpoint, asset_manager_id)
        response = self.session.put(url, json=mtm_result_json)
        if response.ok:
            mtm_results = []
            for mtm_result_json in response.json():
                mtm_results.append(json_to_mtm_result(mtm_result_json))
            return mtm_results
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_mtm_result(self, book_id, asset_manager_id, paramaters):
        """
        parameters is a dictionary of all the mtm result filter parameters
        """
        self.logger.info('Retrieving mtm Positions - Asset Manager: %s', asset_manager_id)
        url = '%s/mtm/%s' % (self.endpoint, asset_manager_id)       
        paramaters.update({'book_id': book_id})  
        response = self.session.get(url, params = paramaters)
        if response.ok:
            mtm_results = [json_to_mtm_result(json_mtm_result) for json_mtm_result in response.json()]
            self.logger.info('Returned %s mtm results.', len(mtm_results))
            return mtm_results
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def position_search(self, asset_manager_id, book_ids=None, account_ids=None,
                        accounting_types=['Transaction Date'], asset_ids=None, position_date=None):
        self.logger.info('Search Positions - Asset Manager: %s', asset_manager_id)
        search_params = {}
        # Potentially roll into a loop
        if book_ids:
            search_params['book_ids'] = ','.join(book_ids)
        if account_ids:
            search_params['account_ids'] = ','.join(account_ids)
        if accounting_types:
            search_params['accounting_types'] = ','.join(accounting_types)
        if asset_ids:
            search_params['asset_ids'] = ','.join(asset_ids)
        if position_date:
            search_params['position_date'] = position_date
        url = '%s/positions/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url, params=search_params)
        if response.ok:
            positions = [json_to_position(json_position) for json_position in response.json()]
            self.logger.info('Returned %s Positions.', len(positions))
            return positions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    # Should this method just be collapsed into positions_by_asset_manager?
    def positions_by_asset_manager_book(self, asset_manager_id, book_id):
        self.logger.info('Retrieve Positions by Asset Manager: %s and Book: %s', asset_manager_id, book_id)
        url = '%s/positions/%s/%s' % (self.endpoint, asset_manager_id, book_id)
        response = self.session.get(url)
        if response.ok:
            positions = [json_to_position(json_position) for json_position in response.json()]
            self.logger.info('Returned %s Positions.', len(positions))
            return positions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def positions_by_asset_manager(self, asset_manager_id, book_ids=None):
        self.logger.info('Retrieve Positions by Asset Manager: %s', asset_manager_id)
        url = '%s/positions/%s' % (self.endpoint, asset_manager_id)
        params = {'book_ids': ','.join(book_ids)} if book_ids else {}
        response = self.session.get(url, params=params)
        if response.ok:
            positions = [json_to_position(json_position) for json_position in response.json()]
            self.logger.info('Returned %s Positions.', len(positions))
            return positions
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def allocate_transaction(self, asset_manager_id, transaction_id, allocation_type, allocation_dicts):
        """

        :param asset_manager_id:
        :param transaction_id:
        :param allocation_type:
        :param allocation_dicts:
        :return:
        """
        self.logger.info('Allocate Transaction - Asset Manager: %s - Transaction ID: %s', asset_manager_id,
                         transaction_id)
        url = '%s/allocations/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        params = {'allocation_type': allocation_type}
        response = self.session.post(url, params=params, data=json.dumps(allocation_dicts, default=json_handler),
                                     headers=self.json_header)
        if response.ok:
            allocations = [json_to_transaction(json_allocation) for json_allocation in response.json()]
            allocation_ids = [allocation.transaction_id for allocation in allocations]
            self.logger.info('%s Allocations Created - Transactions: %s', len(allocations), allocation_ids)
            return allocations
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_transaction_allocations(self, asset_manager_id, transaction_id):
        """

        :param asset_manager_id:
        :param transaction_id:
        :return:
        """
        self.logger.info('Retrieve Allocations - Asset Manager: %s - Transaction ID: %s', asset_manager_id,
                         transaction_id)
        url = '%s/allocations/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        response = self.session.get(url)
        if response.ok:
            allocations = [json_to_transaction(json_allocation) for json_allocation in response.json()]
            self.logger.info('Returned %s Allocations.', len(allocations))
            return allocations
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def net_transactions(self, asset_manager_id, transaction_ids, netting_type='Net'):
        """

        :param asset_manager_id: The asset_manager_id of the netting set owner
        :param transaction_ids:  A list of transaction_ids to net
        :param netting_type:
        :return:
        """
        self.logger.info('Net Transactions - Asset Manager: %s - Transaction IDs: %s', asset_manager_id,
                         transaction_ids)
        url = '%s/netting/%s' % (self.endpoint, asset_manager_id)
        params = {'netting_type': netting_type}
        response = self.session.post(url, params=params, json=transaction_ids)
        if response.ok:
            net_transaction = json_to_transaction(response.json())
            self.logger.info('Net Created - Transaction: %s', net_transaction.transaction_id)
            return net_transaction
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve_netting_set(self, asset_manager_id, transaction_id):
        """
        Returns all the transaction_ids associated with a single netting set.  Pass in the ID for any transaction in
        the set.
        :param asset_manager_id:  The asset_manager_id for the netting set owner.
        :param transaction_id: A transaction_id of an entry within the netting set.
        :return:
        """
        self.logger.info('Retrieve Netting Set - Asset Manager: %s - Transaction ID: %s', asset_manager_id,
                         transaction_id)
        url = '%s/netting/%s/%s' % (self.endpoint, asset_manager_id, transaction_id)
        response = self.session.get(url)
        if response.ok:
            net_transaction_id, netting_set_json = next(iter(response.json().items()))
            netting_set = [json_to_transaction(net_transaction) for net_transaction in netting_set_json]
            self.logger.info('Returned %s Transactions in Netting Set.', len(netting_set))
            return net_transaction_id, netting_set
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def upsert_transaction_asset(self, transaction_asset_json):
        """
        This API should not be called in normal circumstances as the asset cache will populate itself from the assets
        which are created via the Assets API.  However, it can be useful for certain testing scenarios.
        :param transaction_asset_json:
        :return:
        """
        url = self.endpoint + '/assets'
        response = self.session.post(url, json=transaction_asset_json)
        if response.ok:
            return response.json()
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def upsert_transaction_book(self, transaction_book_json):
        """
        This API should not be called in normal circumstances as the book cache will populate itself from the book
        which are created via the Books API.  However, it can be useful for certain testing scenarios.
        :param transaction_book_json:
        :return:
        """
        url = self.endpoint + '/books'
        response = self.session.post(url, json=transaction_book_json)
        if response.ok:
            return response.json()
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def book_transfer(self, asset_manager_id, asset_id, source_book_id, target_book_id, wash_book_id, quantity, price,
                      currency):
        """
        A method for moving between books.  The convention is always *from* the source, *to* the target.

        Two transactions are booked -  one against each

        :param asset_manager_id: The owning asset manager id.
        :param asset_id: The asset being transferred.
        :param source_book_id: The book id of the source.
        :param target_book_id:  The book id of the target.
        :param wash_book_id:  The book id of the wash book which will be the counterparty for the two sides.
        :param quantity: The quantity to transfer.
        :param price:  The price at which to make the transfer.  Typically T-1 EOD price or current market price.
        :param currency: The currency for the transfer price.
        :return:
        """
        url = '%s/book_transfer/%s' % (self.endpoint, asset_manager_id)
        body = {'asset_id': asset_id, 'source_book_id': source_book_id, 'target_book_id': target_book_id,
                'wash_book_id': wash_book_id, 'quantity': quantity, 'price': price, 'currency': currency}
        response = self.session.post(url, data=json.dumps(body, default=json_handler), headers=self.json_header)
        if response.ok:
            deliver_json, receive_json = response.json()
            return json_to_transaction(deliver_json), json_to_transaction(receive_json),
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def depot_transfer(self, asset_manager_id, asset_id, source_account_id, target_account_id, quantity):
        raise NotImplementedError("""This is not yet supported.  The concept is similar to a
                                  book transfer, except it requires an external message to a
                                  custodian to instruct them to move the stock to a
                                   different depot account.""")

    def clear(self, asset_manager_id, book_ids=None):
        """ This method deletes all the data for an asset_manager_id
            and option book_ids.
            It should be used with extreme caution.  In production it
            is almost always better to Inactivate rather than delete. """
        self.logger.info('Clear Transactions & Positions - Asset Manager: %s', asset_manager_id)
        url = '%s/clear/%s' % (self.endpoint, asset_manager_id)
        params = {'asset_manager_ids': ','.join(book_ids)} if book_ids else {}
        response = self.session.delete(url, params=params)
        if response.ok:
            tran_count = response.json().get('transaction_count', 'Unknown')
            self.logger.info('Deleted %s Transactions.', tran_count)
            pos_count = response.json().get('position_count', 'Unknown')
            self.logger.info('Deleted %s Positions.', pos_count)
            return response.json()
        else:
            self.logger.error(response.text)
            response.raise_for_status()


from decimal import Decimal


class OwnershipMixin(object):

    @property
    def ownership(self):
        if hasattr(self, '_ownership'):
            return self._ownership

    @ownership.setter
    def ownership(self, ownership):
        """
        A list of dictionaries in format {'party_id': 'XYZ', 'split': Decimal('0.5')}
        :param ownership:
        :return:
        """
        error_msg = 'ownership must be a list of dictionaries'
        if ownership:
            if not isinstance(ownership, list):
                raise TypeError(error_msg)
            if not all([isinstance(item, dict) for item in ownership]):
                raise TypeError(error_msg)
            if not all([item.get('party_id') for item in ownership]):
                raise ValueError('Ownership missing one or more party_ids')
            if not all([item.get('split') for item in ownership]):
                raise ValueError('Ownership missing one or more splits')
            if sum([item.get('split') for item in ownership]) != Decimal('1'):
                raise ValueError('Ownership must sum up to 100%')
        self._ownership = ownership



from __future__ import absolute_import, division, print_function, unicode_literals

from amaasutils.random_utils import random_string
import random

from amaascore.asset_managers.asset_manager import AssetManager
from amaascore.asset_managers.eod_book import EODBook
from amaascore.asset_managers.enums import ASSET_MANAGER_TYPES, RELATIONSHIP_TYPES
from amaascore.asset_managers.relationship import Relationship


def generate_asset_manager(asset_manager_id=None, asset_manager_type=None, client_id=None,
                           asset_manager_status='Active'):
    asset_manager = AssetManager(asset_manager_id=asset_manager_id,
                                 asset_manager_type=asset_manager_type or random.choice(list(ASSET_MANAGER_TYPES)),
                                 client_id=client_id or random.randint(1, 2**31-1),
                                 asset_manager_status=asset_manager_status)
    return asset_manager


def generate_relationship(asset_manager_id=None, client_id=None, related_id=None, relationship_type=None):
    relationship = Relationship(asset_manager_id=asset_manager_id,
                                relationship_id=str(random.randint(1, 2**31-1)),
                                client_id=client_id or random.randint(1, 2**31-1),
                                relationship_type=relationship_type or random.choice(list(RELATIONSHIP_TYPES)),
                                related_id=related_id or random.randint(1, 2**31-1))
    return relationship

def generate_eod_book(asset_manager_id=None, book_id=None, utc_close_time='18:00:00'):
    eod_book = EODBook(asset_manager_id=asset_manager_id, 
                       book_id=book_id or str(random.randint(1, 2**31-1)),
                       utc_close_time=utc_close_time)
    return eod_book



from decimal import Decimal
from dateutil.parser import parse

import pytz
from amaascore.core.amaas_model import AMaaSModel


class MTMResult(AMaaSModel):

    @staticmethod
    def stored_attributes():
        return {'asset_manager_id', 'book_id', 'mtm_value', 'asset_id',
                'business_date', 'mtm_timestamp', 'message', 'version',
                'mtm_status'}

    @staticmethod
    def lambda_calculated_attributes():
        """
        These fields are calculated internally.  Do not set manually.
        If set manually they will be overridden when the transaction
        is persisted.
        :return:
        """
        return {'client_id', 'created_by', 'updated_by'}


    def __init__(self, asset_manager_id, book_id, business_date, mtm_timestamp, asset_id, 
                 message='', client_id=None, mtm_value=None, mtm_status='Active', *args, **kwargs):
        self.client_id=client_id
        self.asset_manager_id=asset_manager_id 
        self.asset_id=asset_id
        self.book_id=book_id
        self.mtm_value=mtm_value
        self.business_date=business_date
        self.mtm_timestamp=mtm_timestamp 
        self.message=message
        self.mtm_status=mtm_status

        super(MTMResult, self).__init__(*args, **kwargs)

    @property
    def mtm_timestamp(self):
        return self._mtm_timestamp

    @mtm_timestamp.setter
    def mtm_timestamp(self, value):
        if value:
            mtm_timestamp = parse(value) if isinstance(value, str) else value
            self._mtm_timestamp = mtm_timestamp.replace(tzinfo=pytz.UTC) \
                if not mtm_timestamp.tzinfo else mtm_timestamp

    @property
    def mtm_value(self):
        return self._mtm_value

    @mtm_value.setter
    def mtm_value(self, val):
        if not isinstance(val, Decimal) and val is not None:
            self._mtm_value = Decimal(val)
        else:
            self._mtm_value = val


from __future__ import absolute_import, division, print_function, unicode_literals

from dateutil.parser import parse
import logging

from amaascore.config import ENVIRONMENT
from amaascore.core.interface import Interface


class FundamentalsInterface(Interface):

    def __init__(self, environment=ENVIRONMENT, logger=None, endpoint=None, username=None, password=None):
        logger = logger or logging.getLogger(__name__)
        super(FundamentalsInterface, self).__init__(endpoint=endpoint,
                                                    endpoint_type='fundamentals',
                                                    environment=environment,
                                                    username=None, 
                                                    password=None,
                                                    logger=logger)

    def countries(self, country_code=None):
        log_msg = 'Get Country: %s' % country_code if country_code else 'Get All Countries'
        self.logger.info(log_msg)
        url = '%s/countries' % self.endpoint
        params = {'country_code': country_code} if country_code else {}
        response = self.session.get(url, params=params)
        if response.ok:
            self.logger.info('Successfully retrieved country(s)')
            return response.json()
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def holidays(self, country_codes=[], years=[]):
        country_codes = ','.join(country_codes)
        years = ','.join([str(year) for year in years])
        self.logger.info('Get Holiday Calendar for: %s for years: %s', country_codes, years)
        url = '%s/holidays' % self.endpoint
        params = {'country_codes': country_codes,
                  'years': years}

        response = self.session.get(url, params=params)
        if response.ok:
            self.logger.info('Successfully retrieved holidays')
            return response.json()
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def calc_business_date(self, start_date, country_codes, offset, invalid_dates=None):
        self.logger.info('Calculating business date')
        url = '%s/business-date' % self.endpoint
        params = {'start_date': start_date.isoformat(),
                  'country_codes': ','.join(country_codes),
                  'offset': offset}
        if invalid_dates:
            params['invalid_dates'] = ','.join([invalid_date.isoformat() for invalid_date in invalid_dates])
        response = self.session.get(url, params=params)
        if response.ok:
            self.logger.info('Successfully calculated business date')
            business_date = response.json().get('business_date')
            business_date = parse(business_date).date()
            return business_date
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def get_date_info(self, business_date, country_codes):
        self.logger.info('Getting information about date: %s', business_date)
        url = '%s/date-info/%s' % (self.endpoint, business_date.isoformat())
        params = {'country_codes': ','.join(country_codes)}
        response = self.session.get(url, params=params)
        if response.ok:
            self.logger.info('Successfully got information about date')
            return response.json()
        else:
            self.logger.error(response.text)
            response.raise_for_status()


from __future__ import absolute_import, division, print_function, unicode_literals

PARTY_STATUSES = {'Active', 'Inactive', 'Superseded'}
PARTY_TYPES = {'AssetManager', 'Broker', 'Company', 'Exchange', 'Fund', 'GovernmentAgency',
               'Individual', 'Organisation', 'Party'}




from __future__ import absolute_import, division, print_function, unicode_literals


from amaascore.assets.asset import Asset


class CustomAsset(Asset):

    def __init__(self, asset_manager_id, asset_id, client_additional, maturity_date=None, asset_issuer_id=None,
                 asset_status='Active', roll_price=False, display_name='', description='', country_id=None,
                 currency=None, venue_id=None, issue_date=None, fungible=False,
                 links=None, references=None,
                 *args, **kwargs):
        self.maturity_date = maturity_date
        super(CustomAsset, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id, fungible=fungible,
                                          asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                          roll_price=roll_price, display_name=display_name, currency=currency,
                                          description=description, country_id=country_id, venue_id=venue_id,
                                          links=links, references=references,
                                          issue_date=issue_date, client_additional=client_additional,
                                          *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime, date
from dateutil import parser

from amaascore.assets.fund import Fund


class ExchangeTradedFund(Fund):

    def __init__(self, asset_manager_id, asset_id, nav=None, expense_ratio=None, net_assets=None,
                 asset_issuer_id=None, asset_status='Active', description='', display_name='', roll_price=False,
                 country_id=None, venue_id=None, currency=None, creation_date=None,
                 links=None, references=None,
                 *args, **kwargs):
        super(ExchangeTradedFund, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                                 asset_issuer_id=asset_issuer_id,
                                                 asset_status=asset_status, roll_price=roll_price,
                                                 display_name=display_name, description=description,
                                                 country_id=country_id, venue_id=venue_id,
                                                 creation_date=creation_date, fund_type='ETF',
                                                 net_assets=net_assets, nav=nav, expense_ratio=expense_ratio,
                                                 currency=currency, links=links, references=references,
                                                 *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from dateutil.parser import parse
import sys

from amaascore.parties.party import Party

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class Individual(Party):

    def __init__(self, asset_manager_id, party_id, given_names='', surname='', date_of_birth=None, base_currency=None,
                 party_status='Active', description=None, url='', 
                 addresses=None, comments=None, emails=None, links=None, references=None, *args, **kwargs):
        if not hasattr(self, 'party_class'):  # A more specific child class may have already set this
            self.party_class = 'Individual'
        self.given_names = given_names
        self.surname = surname
        self.date_of_birth = date_of_birth
        description = description or '%s, %s' % (self.surname, self.given_names)
        super(Individual, self).__init__(asset_manager_id=asset_manager_id, party_id=party_id,
                                         base_currency=base_currency, party_status=party_status,
                                         description=description, url=url,
                                         addresses=addresses, comments=comments, emails=emails,
                                         links=links, references=references,
                                         *args, **kwargs)

    @property
    def date_of_birth(self):
        if hasattr(self, '_date_of_birth'):
            return self._date_of_birth

    @date_of_birth.setter
    def date_of_birth(self, value):
        """
        The date of birth of the individual.
        :param value:
        :return:
        """
        if value:
            self._date_of_birth = parse(value).date() if isinstance(value, type_check) else value



# POTENTIALLY SUPPORT MULTIPLE ENVIRONMENTS
from __future__ import absolute_import, division, print_function, unicode_literals

envs = {'local', 'dev', 'staging', 'production'}

ENVIRONMENT = 'production'
API_VERSION = 'v1.0'

LOCAL_ENDPOINT = 'http://localhost:8000'
BASE_URLS = {
    'production': 'https://api.amaas.com',
    'staging': 'https://api-staging.amaas.com',
    'dev': 'https://api-dev.amaas.com'
}

ENDPOINTS = {
    'asset_managers': 'assetmanager',
    'assets': 'asset',
    'books': 'book',
    'corporate_actions': 'corporateaction',
    'fundamentals': 'fundamental',
    'market_data': 'marketdata',
    'monitor': 'monitor',
    'parties': 'party',
    'transactions': 'transaction'
}

# Do not change this
COGNITO_REGION = 'ap-southeast-1'
COGNITO_POOL = 'ap-southeast-1_De6j7TWIB'
# This is not secret - it is just an identifier
COGNITO_CLIENT_ID = '2qk35mhjjpk165vssuqhqoi1lk'



from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.core.amaas_model import AMaaSModel


class Comment(AMaaSModel):

    def __init__(self, comment_value, *args, **kwargs):
        self.comment_value = comment_value
        super(Comment, self).__init__(*args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from amaasutils.random_utils import random_string
import datetime
from decimal import Decimal
import random

from amaascore.corporate_actions.corporate_action import CorporateAction
from amaascore.corporate_actions.dividend import Dividend
from amaascore.corporate_actions.split import Split
from amaascore.core.reference import Reference

REFERENCE_TYPES = ['CAEF']


def generate_common(asset_manager_id=None, corporate_action_id=None, asset_id=None):

    common = {'asset_manager_id': asset_manager_id or random.randint(1, 1000),
              'corporate_action_id': corporate_action_id or str(random.randint(1, 10000)),
              'asset_id': asset_id or str(random.randint(1, 10000)),
              'record_date': datetime.date(random.randint(2016, 2020), random.randint(1, 12), random.randint(1, 28))
              }

    return common


def generate_corporate_action(asset_manager_id=None, corporate_action_id=None, asset_id=None):

    common = generate_common(asset_manager_id=asset_manager_id, corporate_action_id=corporate_action_id,
                             asset_id=asset_id)
    corporate_action = CorporateAction(**common)
    references = {ref_type: Reference(reference_value=random_string(10)) for ref_type in REFERENCE_TYPES}
    corporate_action.references.update(references)
    return corporate_action


def generate_dividend(asset_manager_id=None, corporate_action_id=None, asset_id=None):
    attributes = generate_common(asset_manager_id=asset_manager_id, corporate_action_id=corporate_action_id,
                                 asset_id=asset_id)
    dividend = Dividend(dividend_rate=Decimal(random.random()).quantize(Decimal('1.00')),
                        dividend_asset_id=random_string(10), **attributes)
    return dividend


def generate_split(asset_manager_id=None, corporate_action_id=None, asset_id=None):
    attributes = generate_common(asset_manager_id=asset_manager_id, corporate_action_id=corporate_action_id,
                                 asset_id=asset_id)
    split = Split(ratio=(random.randint(1, 5), random.randint(1, 5)), **attributes)
    return split


def generate_corporate_actions(asset_manager_ids=[], number=5):
    corporate_actions = []
    for i in range(number):
        corporate_action = generate_corporate_action(asset_manager_id=random.choice(asset_manager_ids))
        corporate_actions.append(corporate_action)
    return corporate_actions



from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.assets.asset import Asset


class Equity(Asset):

    def __init__(self, asset_manager_id, asset_id, share_class='Common', asset_issuer_id=None, asset_status='Active',
                 issue_date=None, display_name='', description='', country_id=None, venue_id=None, roll_price=False,
                 currency=None, links=None, references=None, *args, **kwargs):
        if not hasattr(self, 'asset_class'):  # A more specific child class may have already set this
            self.asset_class = 'Equity'
        self.share_class = share_class
        super(Equity, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id, fungible=True,
                                     asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                     display_name=display_name, roll_price=roll_price, currency=currency,
                                     description=description, country_id=country_id, venue_id=venue_id,
                                     links=links, references=references, issue_date=issue_date,
                                     *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.transactions.transaction import Transaction


class CashTransaction(Transaction):

    def __init__(self, asset_manager_id, asset_book_id, counterparty_book_id, transaction_action,
                 asset_id, quantity, transaction_date, settlement_date, asset=None, execution_time=None,
                 transaction_type='Trade', transaction_id=None, transaction_status='New',
                 charges=None, codes=None, comments=None, links=None, parties=None,
                 rates=None, references=None, *args, **kwargs):
        """

        :param asset_manager_id:
        :param asset_book_id:
        :param counterparty_book_id:
        :param transaction_action:
        :param asset_id:
        :param quantity:
        :param transaction_date:
        :param settlement_date:
        :param price:
        :param transaction_currency:
        :param settlement_currency: The currency in which the transaction will be settled.  Defaults to the
        transaction_currency if not specified.
        :param asset:
        :param execution_time:
        :param transaction_type:
        :param transaction_id:
        :param transaction_status:
        :param charges:
        :param codes:
        :param comments:
        :param links:
        :param parties:
        :param rates:
        :param references:
        :param args:
        :param kwargs:
        """

        super(CashTransaction, self).__init__(asset_manager_id=asset_manager_id,
                                              asset_book_id=asset_book_id,
                                              counterparty_book_id=counterparty_book_id,
                                              transaction_action=transaction_action,
                                              asset_id=asset_id,
                                              quantity=quantity,
                                              transaction_date=transaction_date,
                                              settlement_date=settlement_date,
                                              price=1,
                                              transaction_currency=asset_id,
                                              settlement_currency=asset_id,
                                              asset=asset,
                                              execution_time=execution_time,
                                              transaction_type=transaction_type,
                                              transaction_id=transaction_id,
                                              transaction_status=transaction_status,
                                              charges=charges,
                                              codes=codes,
                                              comments=comments,
                                              links=links,
                                              parties=parties,
                                              rates=rates,
                                              references=references,
                                              *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
from dateutil.parser import parse
from decimal import Decimal
import sys

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class Quote(object):

    def __init__(self, asset_manager_id, asset_id, quote_datetime, bid=None, ask=None):
        self.asset_manager_id = asset_manager_id
        self.asset_id = asset_id
        self.quote_datetime = quote_datetime
        self.bid = bid
        self.ask = ask

    @property
    def quote_datetime(self):
        return self._quote_datetime

    @quote_datetime.setter
    def quote_datetime(self, value):
        """
        Force the quote_datetime to always be a datetime
        :param value:
        :return:
        """
        if value:
            if isinstance(value, type_check):
                self._quote_datetime = parse(value)
            elif isinstance(value, datetime.datetime):
                self._quote_datetime = value

    @property
    def bid(self):
        return self._bid

    @bid.setter
    def bid(self, value):
        """
        Force the bid to always be a decimal
        :param value:
        :return:
        """
        if value is not None:
            self._bid = Decimal(value)

    @property
    def ask(self):
        return self._ask

    @ask.setter
    def ask(self, value):
        """
        Force the ask to always be a decimal
        :param value:
        :return:
        """
        if value is not None:
            self._ask = Decimal(value)

    def mid(self):
        return (self.bid + self.ask) / 2


from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.asset_managers.asset_manager import AssetManager
from amaascore.asset_managers.relationship import Relationship
from amaascore.asset_managers.eod_book import EODBook

def json_to_asset_manager(json_asset_manager):
    asset_manager = AssetManager(**json_asset_manager)
    return asset_manager


def json_to_relationship(json_relationship):
    relationship = Relationship(**json_relationship)
    return relationship

def json_to_eod_book(json_eod_book):
    eod_book = EODBook(**json_eod_book)
    return eod_book



from __future__ import absolute_import, division, print_function, unicode_literals

from dateutil.relativedelta import relativedelta


class Tenor(object):

    def __init__(self, tenor=None):
        if not self.check_tenor(tenor):
            raise ValueError("Invalid tenor: %s" % tenor)
        self.tenor = tenor

    @staticmethod
    def valid_tenors():
        return ['1M', '3M', '6M', '9M', '1Y', '2Y', '5Y', '10Y', '15Y', '20Y', '30Y', '40Y', '50Y']

    @classmethod
    def check_tenor(cls, tenor):
        return tenor in cls.valid_tenors()

    @classmethod
    def tenor_to_relativedelta(cls, tenor):
        if not cls.check_tenor(tenor):
            raise ValueError("Invalid tenor: %s" % tenor)
        tenor_mapping = {'M': 'months', 'Y': 'years'}
        time_quantity = int(tenor[:-1])
        time_attribute = tenor_mapping.get(tenor[-1])
        relative_delta_attribute = {time_attribute: time_quantity}
        return relativedelta(**relative_delta_attribute)

    def to_relativedelta(self):
        return self.tenor_to_relativedelta(self.tenor)



from __future__ import absolute_import, division, print_function, unicode_literals

import boto3
from configparser import ConfigParser, NoSectionError
from datetime import datetime
import logging
from os.path import expanduser, join
from os import environ
import requests
from warrant.aws_srp import AWSSRP

from amaascore.config import COGNITO_REGION, COGNITO_CLIENT_ID, COGNITO_POOL, ENDPOINTS, LOCAL_ENDPOINT,\
    BASE_URLS, ENVIRONMENT, API_VERSION
from amaascore.exceptions import AMaaSException


class AMaaSSession(object):

    __shared_state = {}

    def __init__(self, username, password, logger):
        if not AMaaSSession.__shared_state:
            AMaaSSession.__shared_state = self.__dict__
            self.refresh_period = 45 * 60  # minutes * seconds
            self.username = username
            self.password = password
            self.tokens = None
            self.last_authenticated = None
            self.session = requests.Session()
            self.client = boto3.client('cognito-idp', COGNITO_REGION)
            self.aws = AWSSRP(username=self.username, password=self.password, pool_id=COGNITO_POOL,
                              client_id=COGNITO_CLIENT_ID, client=self.client)
            self.logger = logger
        else:
            self.__dict__ = AMaaSSession.__shared_state
        if self.needs_refresh():
            self.login()

    def needs_refresh(self):
        if not (self.last_authenticated and
                (datetime.utcnow() - self.last_authenticated).seconds < self.refresh_period):
            return True
        else:
            return False

    def login(self):
        self.logger.info("Attempting login for: %s", self.username)
        try:
            self.tokens = self.aws.authenticate_user().get('AuthenticationResult')
            self.logger.info("Login successful")
            self.last_authenticated = datetime.utcnow()
            self.session.headers.update({'Authorization': self.tokens.get('IdToken')})
        except self.client.exceptions.NotAuthorizedException as e:
            self.logger.info("Login failed")
            self.logger.error(e.response.get('Error'))
            self.last_authenticated = None

    def put(self, url, data=None, **kwargs):
        # Add a refresh
        if self.last_authenticated and not self.needs_refresh():
            return self.session.put(url=url, data=data, **kwargs)
        else:
            raise AMaaSException('Not Authenticated')

    def post(self, url, data=None, **kwargs):
        # Add a refresh
        if self.last_authenticated and not self.needs_refresh():
            return self.session.post(url=url, data=data, **kwargs)
        else:
            raise AMaaSException('Not Authenticated')

    def delete(self, url, **kwargs):
        # Add a refresh
        if self.last_authenticated and not self.needs_refresh():
            return self.session.delete(url=url, **kwargs)
        else:
            raise AMaaSException('Not Authenticated')

    def get(self, url, **kwargs):
        # Add a refresh
        if self.last_authenticated and not self.needs_refresh():
            return self.session.get(url=url, **kwargs)
        else:
            raise AMaaSException('Not Authenticated')

    def patch(self, url, data=None, **kwargs):
        # Add a refresh
        if self.last_authenticated and not self.needs_refresh():
            return self.session.patch(url=url, data=data, **kwargs)
        else:
            raise AMaaSException('Not Authenticated')


class Interface(object):
    """
    Currently this class doesn't do anything - but I anticipate it will be needed in the future.
    """

    def __init__(self, endpoint_type, endpoint=None, environment=ENVIRONMENT, username=None, password=None,
                 config_filename=None, logger=None):
        self.logger = logger or logging.getLogger(__name__)
        self.config_filename = config_filename
        self.endpoint_type = endpoint_type
        self.environment = environment
        self.endpoint = endpoint or self.get_endpoint()
        self.json_header = {'Content-Type': 'application/json'}
        username = username or environ.get('AMAAS_USERNAME') or self.read_config('username')
        password = password or environ.get('AMAAS_PASSWORD') or self.read_config('password')
        self.session = AMaaSSession(username, password, self.logger)
        self.logger.info('Interface Created')

    def get_endpoint(self):
        if self.environment == 'local':
            return LOCAL_ENDPOINT
        if self.environment not in BASE_URLS:
            raise KeyError('Invalid environment specified.')

        base_url = BASE_URLS[self.environment]
        endpoint = ENDPOINTS.get(self.endpoint_type)
        if not endpoint:
            raise KeyError('Cannot find endpoint')
        endpoint = '/'.join([base_url, API_VERSION, endpoint])
        self.logger.info("Using Endpoint: %s", endpoint)
        return endpoint

    @staticmethod
    def generate_config_filename():
        home = expanduser("~")
        return join(home, '.amaas.cfg')

    def read_config(self, option):
        if self.config_filename is None:
            self.config_filename = self.generate_config_filename()
        parser = ConfigParser()
        parser.read(self.config_filename)
        try:
            option = parser.get(section='auth', option=option)
        except NoSectionError:
            raise AMaaSException('Invalid AMaaS config file')
        return option



from __future__ import absolute_import, division, print_function, unicode_literals

import json
import logging

from amaascore.config import ENVIRONMENT, LOCAL_ENDPOINT
from amaascore.core.amaas_model import json_handler
from amaascore.core.interface import Interface
from amaascore.market_data.utils import json_to_eod_price, json_to_fx_rate, json_to_curve


class MarketDataInterface(Interface):

    def __init__(self, environment=ENVIRONMENT, logger=None, endpoint=None, username=None, password=None):
        self.logger = logger or logging.getLogger(__name__)
        super(MarketDataInterface, self).__init__(endpoint=endpoint, endpoint_type='market_data',
                                                  environment=environment, username=username, password=password)

    def persist_eod_prices(self, asset_manager_id, business_date, eod_prices, update_existing_prices=True):
        """

        :param asset_manager_id:
        :param business_date: The business date for which these are rates.  Not really needed, could be derived...
        :param eod_prices:
        :param update_existing_prices:
        :return:
        """
        self.logger.info('Persist EOD Prices - Asset Manager: %s - Business Date: %s', asset_manager_id, business_date)
        url = '%s/eod-prices/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        params = {'update_existing_prices': update_existing_prices}
        eod_prices_json = [eod_price.to_interface() for eod_price in eod_prices]
        response = self.session.post(url, params=params, json=eod_prices_json)
        if response.ok:
            eod_prices = [json_to_eod_price(eod_price) for eod_price in response.json()]
            return eod_prices
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def retrieve_eod_prices(self, asset_manager_id, business_date, asset_ids=None, last_available=False):
        self.logger.info('Retrieve EOD Prices - Asset Manager: %s - Business Date: %s', asset_manager_id, business_date)
        url = '%s/eod-prices/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        params = {'asset_ids': ','.join(asset_ids)} if asset_ids else {}
        # setting last_available to True will return the latest price of the asset as of business 
        # date (or latest prior price if price found on the date)
        if last_available:  
            params.update({'last_available':True})
        response = self.session.get(url=url, params=params)
        if response.ok:
            eod_prices = [json_to_eod_price(eod_price) for eod_price in response.json()]
            self.logger.info('Returned %s EOD Prices.', len(eod_prices))
            return eod_prices
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def roll_prices(self, asset_manager_id, previous_date, asset_ids, update_existing_prices=False):
        url = '%s/roll-prices/%s' % (self.endpoint, asset_manager_id)
        params = {'update_existing_prices': update_existing_prices}
        json_body = json.loads(json.dumps({'previous_date': previous_date,
                                           'asset_ids': ','.join(asset_ids)}, default=json_handler))
        response = self.session.post(url=url, params=params, json=json_body)
        if response.ok:
            prices = [json_to_eod_price(price) for price in response.json()]
            self.logger.info('Rolled %s Prices.', len(prices))
            return prices
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def persist_fx_rates(self, asset_manager_id, business_date, fx_rates, update_existing_rates=True):
        """

        :param asset_manager_id:
        :param business_date: The business date for which these are rates.  Not really needed, could be derived...
        :param fx_rates:
        :param update_existing_rates:
        :return:
        """
        self.logger.info('Persist FX Rates - Asset Manager: %s - Business Date: %s', asset_manager_id, business_date)
        url = '%s/fx-rates/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        params = {'update_existing_rates': update_existing_rates}
        fx_rates_json = [fx_rate.to_interface() for fx_rate in fx_rates]
        response = self.session.post(url, params=params, json=fx_rates_json)
        if response.ok:
            fx_rates = [json_to_fx_rate(fx_rate) for fx_rate in response.json()]
            return fx_rates
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def retrieve_fx_rates(self, asset_manager_id, business_date, asset_ids=None):
        self.logger.info('Retrieve FX Rates - Asset Manager: %s - Business Date: %s', asset_manager_id, business_date)
        url = '%s/fx-rates/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        params = {'asset_ids': ','.join(asset_ids)} if asset_ids else {}
        response = self.session.get(url=url, params=params)
        if response.ok:
            fx_rates = [json_to_fx_rate(fx_rate) for fx_rate in response.json()]
            self.logger.info('Returned %s FX Rates.', len(fx_rates))
            return fx_rates
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def retrieve_curve(self, asset_manager_id, business_date, asset_ids = None):
        self.logger.info('Retrieve curve - Asset Manager: %s - Business Date: %s', asset_manager_id, business_date)
        url = '%s/curves/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        params = {'asset_ids': ','.join(asset_ids)} if asset_ids else {}
        response = self.session.get(url = url, params = params)
        if response.ok:
            curves = [json_to_curve(curve) for curve in response.json() ]
            self.logger.info('Returned %s curves.', len(curves))
            return curves
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def persist_curves(self, asset_manager_id, business_date, curves, update_existing_curves=True):
        """
        :param asset_manager_id:
        :param business_date: 
        :param curves:
        :param update_existing_rates:
        :return:
        """
        self.logger.info('Persist curves - Asset Manager: %s - Business Date: %s', asset_manager_id, business_date)
        url = '%s/curves/%s/%s' % (self.endpoint, asset_manager_id, business_date.isoformat())
        params = {'update_existing_curves': update_existing_curves}
        curves_json = [curve.to_interface() for curve in curves]
        response = self.session.post(url, params = params, json=curves_json)
        if response.ok:
            curves = [json_to_curve(curve) for curve in response.json()]
            return curves
        else:
            self.logger.error(response.text)
            response.raise_for_status()


    def clear(self, asset_manager_id):
        """ This method deletes all the data for an asset_manager_id.
            It should be used with extreme caution.  In production it
            is almost always better to Inactivate rather than delete. """
        self.logger.info('Clear Market Data - Asset Manager: %s', asset_manager_id)
        url = '%s/clear/%s' % (self.endpoint, asset_manager_id)
        response = self.session.delete(url)
        if response.ok:
            eod_price_count = response.json().get('eod_price_count', 'Unknown')
            self.logger.info('Deleted %s EOD Prices.', eod_price_count)
            fx_rate_count = response.json().get('fx_rate_count', 'Unknown')
            self.logger.info('Deleted %s FX Rates.', fx_rate_count)
            return response.json()
        else:
            self.logger.error(response.text)
            response.raise_for_status()
    
    
    def get_brokendate_fx_forward_rate(self, asset_manager_id,  asset_id, price_date, value_date):
        """
        This method takes calculates broken date forward FX rate based on the passed in parameters
        """                
        self.logger.info('Calculate broken date FX Forward - Asset Manager: %s - Asset (currency): %s - Price Date: %s - Value Date: %s', asset_manager_id, asset_id, price_date, value_date)
        url = '%s/brokendateforward/%s' % (self.endpoint, asset_manager_id)
        params = {'value_date': value_date, 'asset_id':asset_id, 'price_date': price_date}
        response = self.session.get(url=url, params = params)
        if response.ok:
            forward_rate = response.json()
            self.logger.info('Retrieved broken date FX forward rate %f', forward_rate)
            return forward_rate
        else:
            self.logger.error(response.text)
            response.raise_for_status()




from __future__ import absolute_import, division, print_function, unicode_literals

from dateutil.parser import parse
import sys
import uuid

from amaascore.core.amaas_model import AMaaSModel

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class Item(AMaaSModel):

    def __init__(self, asset_manager_id, item_class, item_type, item_level, item_source, message, item_id=None,
                 item_status='Open', asset_book_id=None, transaction_id=None, asset_id=None, item_date=None,
                 *args, **kwargs):

        self.asset_manager_id = asset_manager_id
        self.item_id = item_id or uuid.uuid4().hex
        self.item_class = item_class
        self.item_type = item_type
        self.item_level = item_level
        self.item_source = item_source
        self.message = message
        self.item_status = item_status
        self.asset_book_id = asset_book_id
        self.transaction_id = transaction_id
        self.asset_id = asset_id
        self.item_date = item_date
        super(Item, self).__init__(*args, **kwargs)

    @property
    def item_date(self):
        if hasattr(self, '_item_date'):
            return self._item_date

    @item_date.setter
    def item_date(self, item_date):
        """

        :param item_date:
        :return:
        """
        if item_date:
            self._item_date = parse(item_date).date() if isinstance(item_date, type_check) else item_date



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import timedelta
from dateutil.parser import parse

from amaascore.asset_managers.enums import ACCOUNT_TYPES, ASSET_MANAGER_TYPES
from amaascore.core.amaas_model import AMaaSModel
from amaascore.error_messages import ERROR_LOOKUP


class AssetManager(AMaaSModel):

    def __init__(self, asset_manager_type, asset_manager_id=None, asset_manager_status='Active', party_id=None,
                 account_type='Basic', default_book_owner_id='', default_timezone='UTC',
                 default_book_close_time=timedelta(hours=18),
                 *args, **kwargs):
        self.asset_manager_id = asset_manager_id
        self.asset_manager_type = asset_manager_type
        self.asset_manager_status = asset_manager_status
        self.party_id = party_id
        self.default_book_owner_id = default_book_owner_id
        self.default_timezone = default_timezone
        self.default_book_close_time = default_book_close_time
        self.account_type = account_type
        super(AssetManager, self).__init__(*args, **kwargs)

    @property
    def asset_manager_type(self):
        return self._asset_manager_type

    @asset_manager_type.setter
    def asset_manager_type(self, asset_manager_type):
        if asset_manager_type:
            if asset_manager_type in ASSET_MANAGER_TYPES:
                self._asset_manager_type = asset_manager_type
            else:
                raise ValueError(ERROR_LOOKUP.get('am_type_invalid') % (asset_manager_type,
                                                                        self.asset_manager_id))

    @property
    def account_type(self):
        if hasattr(self, '_account_type'):
            return self._account_type

    @account_type.setter
    def account_type(self, account_type):
        if account_type:
            if account_type in ACCOUNT_TYPES:
                self._account_type = account_type
            else:
                raise ValueError(ERROR_LOOKUP.get('am_account_type_invalid') % (account_type,
                                                                                self.asset_manager_id))



import datetime
from amaascore.core.amaas_model import AMaaSModel

class EODBook(AMaaSModel):

    def __init__(self, asset_manager_id, book_id, utc_close_time,
                 eod_book_status='Active', *args, **kwargs):
        self.asset_manager_id = asset_manager_id
        self.book_id = book_id
        self.utc_close_time = utc_close_time.strftime('%H:%M:%S') if isinstance(utc_close_time, datetime.time) else str(utc_close_time)
        self.eod_book_status = eod_book_status
        super(EODBook, self).__init__(*args, **kwargs)


from __future__ import absolute_import, division, print_function, unicode_literals


class AMaaSException(Exception):
    """ Base class for all AMaaS Exceptions """
    pass


class TransactionNeedsSaving(Exception):
    def __init__(self):
        message = "Transaction needs to be saved to AMaaS Core for the functionality to be valid"
        super(TransactionNeedsSaving, self).__init__(message)



from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from amaascore.books.utils import json_to_book
from amaascore.config import ENVIRONMENT
from amaascore.core.interface import Interface


class BooksInterface(Interface):

    def __init__(self, environment=ENVIRONMENT, logger=None, endpoint=None, username=None, password=None):
        logger = logger or logging.getLogger(__name__)
        super(BooksInterface, self).__init__(endpoint=endpoint, endpoint_type='books', environment=environment,
                                             username=username, password=password, logger=logger)

    def new(self, book):
        self.logger.info('New Book - Asset Manager: %s - Book ID: %s', book.asset_manager_id, book.book_id)
        url = '%s/books/%s' % (self.endpoint, book.asset_manager_id)
        response = self.session.post(url, json=book.to_interface())
        if response.ok:
            self.logger.info('Successfully Created Book - Asset Manager: %s - Book ID: %s', book.asset_manager_id,
                             book.book_id)
            book = json_to_book(response.json())
            return book
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def amend(self, book):
        self.logger.info('Amend Book - Asset Manager: %s - Book ID: %s', book.asset_manager_id, book.book_id)
        url = '%s/books/%s/%s' % (self.endpoint, book.asset_manager_id, book.book_id)
        response = self.session.put(url, json=book.to_interface())
        if response.ok:
            self.logger.info('Successfully Amended Book - Asset Manager: %s - Book ID: %s', book.asset_manager_id,
                             book.book_id)
            book = json_to_book(response.json())
            return book
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retrieve(self, asset_manager_id, book_id, version=None):
        self.logger.info('Retrieve Book - Asset Manager: %s - Book ID: %s', asset_manager_id, book_id)
        url = '%s/books/%s/%s' % (self.endpoint, asset_manager_id, book_id)
        if version:
            url += '?version=%d' % int(version)
        response = self.session.get(url)
        if response.ok:
            self.logger.info('Successfully Retrieved Book - Asset Manager: %s - Book ID: %s', asset_manager_id,
                             book_id)
            return json_to_book(response.json())
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def retire(self, asset_manager_id, book_id):
        self.logger.info('Retire Book - Asset Manager: %s - Book ID: %s', asset_manager_id, book_id)
        url = '%s/books/%s/%s' % (self.endpoint, asset_manager_id, book_id)
        json = {'book_status': 'Retired'}
        response = self.session.patch(url, json=json)
        if response.ok:
            self.logger.info('Successfully Retired Book - Asset Manager: %s - Book ID: %s', asset_manager_id, book_id)
            return json_to_book(response.json())
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def search(self, asset_manager_id, book_ids=None, business_units=None, 
                     owner_ids=None, party_ids=None, book_statuses=None):
        self.logger.info('Search Books - Asset Manager: %s', asset_manager_id)
        search_params = {}
        # Potentially roll this into a loop through args rather than explicitly named - depends on additional validation
        if book_ids:
            search_params['book_ids'] = ','.join(book_ids)
        if business_units:
            search_params['business_units'] = ','.join(business_units)
        if owner_ids:
            search_params['owner_ids'] = ','.join(owner_ids)
        if party_ids:
            search_params['party_ids'] = ','.join(party_ids)
        if book_statuses:
            search_params['book_statuses'] = ','.join(book_statuses)
        url = '%s/books/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url, params=search_params)
        if response.ok:
            books = [json_to_book(json_book)
                     for json_book in response.json()] if response.json() else None
            self.logger.info('Returned %s Books.', len(books) if books else 0)
            return books
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def books_by_asset_manager(self, asset_manager_id):
        self.logger.info('Retrieve Books by Asset Manager: %s', asset_manager_id)
        url = '%s/books/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url)
        if response.ok:
            books = [json_to_book(json_book) for json_book in response.json()]
            self.logger.info('Returned %s Books.', len(books))
            return books
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def book_config(self, asset_manager_id):
        self.logger.info('Retrieve Book Config by Asset Manager: %s', asset_manager_id)
        url = '%s/book_config/%s' % (self.endpoint, asset_manager_id)
        response = self.session.get(url)
        if response.ok:
            book_config = response.json()
            self.logger.info('Successfully returned Book Config for %s', asset_manager_id)
            return book_config
        else:
            self.logger.error(response.text)
            response.raise_for_status()

    def clear(self, asset_manager_id):
        """ This method deletes all the data for an asset_manager_id.
            It should be used with extreme caution.  In production it
            is almost always better to Retire rather than delete. """
        self.logger.info('Clear Books - Asset Manager: %s', asset_manager_id)
        url = '%s/clear/%s' % (self.endpoint, asset_manager_id)
        response = self.session.delete(url)
        if response.ok:
            count = response.json().get('count', 'Unknown')
            self.logger.info('Deleted %s Books.', count)
            return count
        else:
            self.logger.error(response.text)
            response.raise_for_status()



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import date
from dateutil.parser import parse
import re
import sys
import uuid
import pickle

from amaascore.assets.children import Link
from amaascore.core.amaas_model import AMaaSModel
from amaascore.core.comment import Comment
from amaascore.core.reference import Reference
from amaasutils.hash import compute_hash


# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class Asset(AMaaSModel):

    @staticmethod
    def children():
        return {'comments': Comment, 'links': Link, 'references': Reference}

    def __init__(self, asset_manager_id, fungible, asset_issuer_id=None, asset_id=None, asset_status='Active',
                 country_id=None, venue_id=None, currency=None, issue_date=date.min,
                 roll_price=True, display_name='', description='',
                 comments=None, links=None, references=None, client_additional=None,
                 *args, **kwargs):
        self.asset_manager_id = asset_manager_id
        self.asset_id = asset_id or uuid.uuid4().hex
        if not hasattr(self, 'asset_class'):  # A more specific child class may have already set this
            self.asset_class = 'Asset'
        # This can be overridden if necessary but I think it probably shouldn't
        self.asset_type_display = ' '.join(re.findall('[A-Z][^A-Z]*', self.__class__.__name__))
        self.asset_type = self.__class__.__name__
        self.fungible = fungible
        self.asset_issuer_id = asset_issuer_id
        self.asset_status = asset_status
        self.country_id = country_id
        self.venue_id = venue_id
        self.currency = currency
        self.issue_date = issue_date
        self.roll_price = roll_price
        self.display_name = display_name
        self.description = description
        self.client_additional = client_additional  # A field to allow people to build their own assets
        # Defaults are here not in constructor for mutability reasons.
        self.comments = comments.copy() if comments else {}
        self.links = links.copy() if links else {}
        self.references = references.copy() if references else {}
        self.references['Argomi'] = Reference(reference_value=self.asset_id)  # Upserts the Argomi Reference
        super(Asset, self).__init__(*args, **kwargs)

    def reference_types(self):
        """
        TODO - are these helper functions useful?
        :return:
        """
        return self.references.keys()

    @property
    def issue_date(self):
        if hasattr(self, '_issue_date'):
            return self._issue_date

    @issue_date.setter
    def issue_date(self, value):
        """
        The date on which the asset is issued
        :param value:
        :return:
        """
        self._issue_date = parse(value).date() if isinstance(value, type_check) else value

    @property
    def maturity_date(self):
        if hasattr(self, '_maturity_date'):
            return self._maturity_date

    @maturity_date.setter
    def maturity_date(self, value):
        """
        The date on which the asset matures and no longer holds value
        :param value:
        :return:
        """
        self._maturity_date = parse(value).date() if isinstance(value, type_check) else value

    def __str__(self):
        return "Asset object - ID: %s" % self.asset_id

    def get_country_codes(self):
        return [self.country_id]

    def country_and_venue_codes(self):
        """
        This function returns dictionary in the format of {"country_codes": [XYZ, ABC], "venue_id": [ABC]} 
        Where country_codes = self.country_id and venue_id = self.venue_id
        except for FX where we also return the country_ids associated with the legs of the FX transaction 
        (e.g. USDJPY returns USA + JPN).
        """
        codes = {'country_codes': self.get_country_codes(), 'venue_id': [self.venue_id]}
        return codes

    def get_currencies(self):
        return [self.currency]

    @property
    def hashcode(self):
        ignored_attributes = ['version']
        boolean_attributes = ['fungible', 'roll_price']
        attributes = {attr: getattr(self, attr) for attr in self.to_dict()
                      if attr not in boolean_attributes}

        for boolean_attr in boolean_attributes:
            attributes[boolean_attr] = int(getattr(self, boolean_attr))

        for attr in ['client_additional']:
            value = getattr(self, attr)
            if value:
                dict_value = pickle.loads(value)
                if dict_value and isinstance(dict_value, dict):
                    attributes[attr] = dict_value
        return compute_hash(attributes, ignored_attributes=ignored_attributes)



from setuptools import setup, find_packages

requires = [
    'amaasutils',
    'configparser',
    'python-dateutil',
    'pytz',
    'requests',
    'warrant'
]

setup(
    name='amaascore',
    version='0.5.22',
    description='Asset Management as a Service - Core SDK',
    license='Apache License 2.0',
    url='https://github.com/amaas-fintech/amaas-core-sdk-python',
    author='AMaaS',
    author_email='tech@amaas.com',
    classifiers=[
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(exclude=['tests']),  # Very annoying that this doesnt work - I have to include a MANIFEST
    install_requires=requires,
)





import logging.config
import csv
import json

from amaascore.tools.csv_tools import csv_stream_to_objects
from amaasutils.logging_utils import DEFAULT_LOGGING
from amaascore.csv_upload.utils import process_normal, interface_direct_class, interface_direct_csvpath

from amaascore.assets.asset import Asset
from amaascore.assets.automobile import Automobile
from amaascore.assets.bond import BondCorporate, BondGovernment, BondMortgage
from amaascore.assets.bond_future import BondFuture
from amaascore.assets.bond_future_option import BondFutureOption
from amaascore.assets.bond_option import BondOption
from amaascore.assets.commodity_future import CommodityFuture
from amaascore.assets.cfd import ContractForDifference
from amaascore.assets.currency import Currency
from amaascore.assets.custom_asset import CustomAsset
from amaascore.assets.derivative import Derivative
from amaascore.assets.energy_future import EnergyFuture
from amaascore.assets.equity import Equity
from amaascore.assets.equity_future import EquityFuture
from amaascore.assets.etf import ExchangeTradedFund
from amaascore.assets.foreign_exchange import ForeignExchange, NonDeliverableForward
from amaascore.assets.fund import Fund
from amaascore.assets.future import Future
from amaascore.assets.future_option import FutureOption
from amaascore.assets.fx_future import ForeignExchangeFuture
from amaascore.assets.fx_option import ForeignExchangeOption
from amaascore.assets.index import Index
from amaascore.assets.index_future import IndexFuture
from amaascore.assets.interest_rate_future import InterestRateFuture
from amaascore.assets.listed_cfd import ListedContractForDifference
from amaascore.assets.listed_derivative import ListedDerivative
from amaascore.assets.option_mixin import OptionMixin
from amaascore.assets.real_asset import RealAsset
from amaascore.assets.real_estate import RealEstate
from amaascore.assets.sukuk import Sukuk
from amaascore.assets.synthetic import Synthetic
from amaascore.assets.synthetic_from_book import SyntheticFromBook
from amaascore.assets.synthetic_multi_leg import SyntheticMultiLeg
from amaascore.assets.wine import Wine
from amaascore.assets.warrants import Warrant

from amaascore.parties.broker import Broker
from amaascore.parties.company import Company
from amaascore.parties.exchange import Exchange
from amaascore.parties.fund import Fund
from amaascore.parties.government_agency import GovernmentAgency
from amaascore.parties.individual import Individual
from amaascore.parties.organisation import Organisation
from amaascore.parties.party import Party
from amaascore.parties.sub_fund import SubFund

from amaascore.books.book import Book

from amaascore.corporate_actions.corporate_action import CorporateAction
from amaascore.corporate_actions.dividend import Dividend
from amaascore.corporate_actions.notification import Notification
from amaascore.corporate_actions.split import Split

from amaascore.market_data.eod_price import EODPrice
from amaascore.market_data.fx_rate import FXRate
from amaascore.market_data.quote import Quote

from amaascore.transactions.position import Position
from amaascore.transactions.transaction import Transaction

from amaascore.asset_managers.asset_manager import AssetManager
from amaascore.asset_managers.relationship import Relationship

class Uploader(object):

    def __init__(self):
        pass

    @staticmethod
    def json_handler(orderedDict, params):
        Dict = dict(orderedDict)
        for key, var in params.items():
            Dict[key]=var
        data_class = Dict.get('amaasclass', None)
        Dict = process_normal(Dict)
        obj = globals()[data_class](**dict(Dict))
        return obj

    @staticmethod
    def upload(csvpath, asset_manager_id=None, client_id=None):
        """convert csv file rows to objects and insert;
           asset_manager_id and possibly client_id from the UI (login)"""
        interface = interface_direct_csvpath(csvpath)
        logging.config.dictConfig(DEFAULT_LOGGING)
        logger = logging.getLogger(__name__)
        if asset_manager_id is None:
            params = dict()
        elif client_id is None:
            params = {'asset_manager_id': asset_manager_id}
        else:
            params = {'asset_manager_id': asset_manager_id, 'client_id': client_id}
        with open(csvpath) as csvfile:
            objs = csv_stream_to_objects(stream=csvfile, json_handler=Uploader.json_handler, params=params)
        for obj in objs:
            interface.new(obj)
            logger.info('Creating this object and upload to database successfully')

    @staticmethod
    def download(csvpath, asset_manager_id, data_id_type, data_id_list):
        """retrieve the objs mainly for test purposes"""
        interface = interface_direct_csvpath(csvpath)
        logging.config.dictConfig(DEFAULT_LOGGING)
        logger = logging.getLogger(__name__)
        objs = []
        for data_id in data_id_list:
            Dict = dict()
            Dict[data_id_type] = data_id
            objs.append(interface.retrieve(asset_manager_id=asset_manager_id, **Dict))
        return objs


from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.assets.asset import Asset


class Sukuk(Asset):

    def __init__(self, asset_manager_id, asset_id, maturity_date, asset_issuer_id=None,
                 asset_status='Active', roll_price=True, issue_date=None, display_name='', description='',
                 country_id=None, venue_id=None, currency=None,
                 comments=None, links=None, references=None, *args, **kwargs):
        if not hasattr(self, 'asset_class'):  # A more specific child class may have already set this
            self.asset_class = 'Sukuk'
        self.maturity_date = maturity_date
        super(Sukuk, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id, fungible=True,
                                    asset_issuer_id=asset_issuer_id, asset_status=asset_status,
                                    roll_price=roll_price, display_name=display_name, currency=currency,
                                    description=description, country_id=country_id, venue_id=venue_id,
                                    comments=comments, links=links, references=references,
                                    issue_date=issue_date,
                                    *args, **kwargs)



from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import date
from dateutil.parser import parse
import sys

from amaascore.assets.derivative import Derivative
from amaascore.assets.option_mixin import OptionMixin

# This extremely ugly hack is due to the whole Python 2 vs 3 debacle.
type_check = str if sys.version_info >= (3, 0, 0) else (str, unicode)


class ForeignExchangeOption(Derivative, OptionMixin):
    """
    An over the counter Option with an underlying FX pair.
    """

    def __init__(self, asset_manager_id, asset_id, option_type, strike, underlying_asset_id, option_style,
                 issue_date=date.min, asset_status='Active', asset_issuer_id=None, expiry_date=date.max,
                 display_name='', description='', links=None, references=None,
                 *args, **kwargs):
        self.option_type = option_type
        self.strike = strike
        self.underlying_asset_id = underlying_asset_id
        self.option_style = option_style
        self.expiry_date = expiry_date
        super(ForeignExchangeOption, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id,
                                                    asset_issuer_id=asset_issuer_id,
                                                    issue_date=issue_date, asset_status=asset_status,
                                                    display_name=display_name, description=description,
                                                    links=links, references=references, *args, **kwargs)

    @property
    def expiry_date(self):
        if hasattr(self, '_expiry_date'):
            return self._expiry_date

    @expiry_date.setter
    def expiry_date(self, value):
        """
        The date on which the Futures contract expires
        :param expiry_date:
        :return:
        """
        if value:
            self._expiry_date = parse(value).date() if isinstance(value, type_check) else value



from __future__ import absolute_import, division, print_function, unicode_literals

from amaascore.assets.asset import Asset


class CurrencyBase(Asset):

    def __init__(self, asset_id, minor_unit_places, asset_manager_id=0, asset_status='Active', display_name='',
                 description='', country_id=None, *args, **kwargs):
        self.asset_class = 'Currency'
        self.minor_unit_places = minor_unit_places
        super(CurrencyBase, self).__init__(asset_manager_id=asset_manager_id, asset_id=asset_id, fungible=True,
                                           display_name=display_name, asset_class=self.asset_class,
                                           asset_status=asset_status, description=description,
                                           country_id=country_id, venue_id=None, currency=asset_id,
                                           *args, **kwargs)


class Currency(CurrencyBase):

    def __init__(self, asset_id, asset_manager_id=0, deliverable=True, asset_status='Active', major=False,
                 minor_unit_places=2, display_name='', description='', country_id=None, *args, **kwargs):
        self.deliverable = deliverable
        self.major = major
        super(Currency, self).__init__(asset_manager_id=asset_manager_id, minor_unit_places=minor_unit_places,
                                       asset_id=asset_id, display_name=display_name, asset_status=asset_status,
                                       description=description, country_id=country_id,
                                       *args, **kwargs)

    @property
    def major(self):
        return self._major

    @major.setter
    def major(self, major):
        """

        :param major:
        :return:
        """
        if major:
            self._major = major
        else:
            self._major = False
